(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Xml imports ProofChecker 
begin

datatype xml = XML string "(string \<times> string) list" "xml list" "string option" (* name, attributes, children, text *)
datatype xmldoc = XMLDOC "string" xml (* header, body *)

definition wspace :: "string" where "wspace = newline @ tab @ '' ''"
definition dquote :: "string" where "dquote = ''\"''"
definition letters :: "string" where "letters = ''abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_0123456789&;:''"

       
fun trim :: "string \<Rightarrow> string" 
  where "trim (c # s) = (if c mem wspace then trim s else c # s)"
     |  "trim [] = []"


fun splitterFun :: "(char \<Rightarrow> bool) \<Rightarrow> string \<Rightarrow> string \<times> string"
  where "splitterFun allowed (c # s) = 
  (if allowed c then (let (l,r) = splitterFun allowed s in (c # l, r)) else ([], c # s))"
     |  "splitterFun _ [] = ([],[])"

fun splitter :: "string \<Rightarrow> string \<Rightarrow> string \<times> string"
  where "splitter allowed s = splitterFun (\<lambda> c. c mem allowed) s" 

types 'a parser = "string \<Rightarrow> (string,'a \<times> string)emonad"
types 'a endParser = "string \<Rightarrow> (string,'a)emonad"

definition eofErr :: "(string,'a)emonad" 
  where "eofErr = error ''unexpected end of input''"

fun parseError :: "string \<Rightarrow> string \<Rightarrow> (string,'a)emonad"
  where "parseError s allowed = error (''expected symbols '' @ allowed @ '' but found '' @ take 30 s)"

fun parseName :: "string parser" 
  where "parseName s = (let (l,r) = splitter letters s in (if l = [] then parseError s letters else return (l,trim r)))" 

fun parseText :: "string option parser" 
  where "parseText s = (let (l,r) = splitterFun (\<lambda> c. [c] \<noteq> ''<'') s in (if trim l = [] then return (None,r) else return (Some (List.rev (trim (List.rev (trim l)))),r)))"

fun consume2 :: "string list \<Rightarrow> string list \<Rightarrow> string parser"
  where "consume2 [] allowed s = parseError s (concat (map (\<lambda> s. s @ '','') allowed))"
     |  "consume2 (t # as) allowed s = (if (map snd (zip t s)) = t then return (t,trim (List.drop (length t) s)) else consume2 as allowed s)"

fun consume :: "string list \<Rightarrow> string parser"
  where "consume allowed s = consume2 allowed allowed s"

fun scanUpto :: "string \<Rightarrow> string parser"
  where "scanUpto end [] = error (''did not find required token '' @ end)"
      | "scanUpto end (c # s) = (if map snd (zip end (c # s)) = end 
             then return (take (length end) (c # s), trim (List.drop (length end) (c # s)))
             else scanUpto end s \<guillemotright>= (\<lambda> (res,rest). return (c # res, rest)))"

fun parseAttribute :: "string parser"
  where "parseAttribute s = (consume [dquote] s \<guillemotright>= 
            (\<lambda> (_,s1). let (n,s2) = splitterFun (\<lambda> c. [c] \<noteq> dquote) s1 in 
                  consume [dquote] s2 \<guillemotright>= (\<lambda> (_,s3). return (n,s3))))" 

lemma bindCong [fundef_cong] : fixes m1 m2 :: "('e,'a)emonad"
  assumes "m1 = m2" and "n1 = n2"
  shows "(m1 \<guillemotright>= n1) = (m2 \<guillemotright>= n2)"
using assms
 by (unfold "bind_def", auto)

function parseAttributes :: "(string \<times> string) list parser"
  where "parseAttributes [] = return ([],[])"
      | "parseAttributes (c # s) = (if c mem ''/>'' then return ([],c # s) else 
                        parseName (c # s) \<guillemotright>= 
                   (\<lambda> (n,s1). consume [''=''] s1 \<guillemotright>=
                   (\<lambda> (_,s2). parseAttribute s2 \<guillemotright>= 
                   (\<lambda> (a,s3). parseAttributes s3 \<guillemotright>=
                   (\<lambda> (nas,s4). return ((n,a) # nas,s4))))))" 
by (pat_completeness, auto)

function parseXMLList :: "xml list parser"
where "parseXMLList s = (if (s = [] \<or> take 2 s = ''</'') then return ([],s) else 
 (consume [''<''] s 
  \<guillemotright>= (\<lambda> (_,s1). parseName s1 
  \<guillemotright>= (\<lambda> (n,s2). parseAttributes s2
  \<guillemotright>= (\<lambda> (nas,s3). consume [''/>'',''>''] s3
  \<guillemotright>= (\<lambda> (clo,s4). if clo = ''/>'' then parseXMLList s4 \<guillemotright>= (\<lambda> (li, s5). return (XML n nas [] None # li, s5)) else 
    parseText s4 
  \<guillemotright>= (\<lambda> (t,s5). parseXMLList s5
  \<guillemotright>= (\<lambda> (chi,s6). consume [''</'' @ n @ ''>''] s6
  \<guillemotright>= (\<lambda> (_,s7). parseXMLList s7 
  \<guillemotright>= (\<lambda> (li,s8). return (XML n nas chi t # li, s8) 
  ))))))))))" 
by (pat_completeness, auto)

fun parseXML :: "xml parser"
where "parseXML s = (consume [''<''] s 
  \<guillemotright>= (\<lambda> (_,s1). parseName s1 
  \<guillemotright>= (\<lambda> (n,s2). parseAttributes s2
  \<guillemotright>= (\<lambda> (nas,s3). consume [''/>'',''>''] s3
  \<guillemotright>= (\<lambda> (clo,s4). if clo = ''/>'' then return (XML n nas [] None, trim s4) else 
    parseText s4 
  \<guillemotright>= (\<lambda> (t,s5). parseXMLList s5
  \<guillemotright>= (\<lambda> (chi,s6). consume [''</'' @ n @ ''>''] s6
  \<guillemotright>= (\<lambda> (_,s7). return (XML n nas chi t, s7)
  ))))))))"

function stripXMLHeader :: "string parser"
where "stripXMLHeader s = (if take 2 (trim s) = ''<?'' 
          then (scanUpto ''?>'' s
                \<guillemotright>= (\<lambda> (header1,s1). stripXMLHeader s1 \<guillemotright>= (\<lambda> (header2, s2). return (header1 +++ header2, trim s2))))
          else return ('''',trim s))"
by (pat_completeness, auto)

fun parseXMLDoc :: "xmldoc endParser"
where "parseXMLDoc s = (stripXMLHeader s
     \<guillemotright>= (\<lambda> (header, rest). parseXML rest \<guillemotright>= (\<lambda> (xml, end). if end = [] then return (XMLDOC header xml) else 
         error (''there is text after root element: '' @ newline @ end))))"

(* and pretty printer *)

instantiation xml :: sshow
begin
fun showAtt :: "string \<times> string \<Rightarrow> string" 
  where "showAtt (a,v) = (a @ ''='' @ dquote @ v @ dquote)"
fun showAtts :: "(string \<times> string) list \<Rightarrow> string"
  where "showAtts as = foldr (\<lambda> a s. '' '' @ showAtt a @ s) as ''''"
fun showXMLindent :: "string \<Rightarrow> nat \<Rightarrow> xml \<Rightarrow> string" 
  where "showXMLindent ind i (XML n a c t) = sConcat [ind,''<'',n,showAtts a,if c = [] \<and> t = None then ''/>'' else 
           sConcat [''>'', (if c \<noteq> [] then (List.append (sConcat (map (showXMLindent (List.append ind (replicate i (CHR '' ''))) i) c)) ind) else []), 
                   (if t = None then [] else the t) , ''</'', n , ''>'']]"
fun sshow_xml where "sshow_xml xml = (trim (showXMLindent newline 2 xml))"
instance ..
end

instantiation xmldoc :: sshow
begin
fun sshow_xmldoc where "sshow_xmldoc (XMLDOC h x) = (h @ sshow x)"
instance ..
end


(* xml to ...  *)

fun mapMonad :: "(xml \<Rightarrow> ('e,'a)emonad) \<Rightarrow> xml list \<Rightarrow> ('e,'a list)emonad"
where "mapMonad _ [] = return []"
    | "mapMonad f (xml # rest) = (case f xml of Left e \<Rightarrow> error e | Right a \<Rightarrow> (case mapMonad f rest of Left e \<Rightarrow> error e | Right as \<Rightarrow> return (a # as)))"

lemma mapMonadCong [fundef_cong] : fixes f1 f2 :: "xml \<Rightarrow> ('e,'a)emonad"
  assumes "f1 = f2" and "l1 = l2"
  shows "(mapMonad f1 l1) = (mapMonad f2 l2)"
using assms by auto

fun xml0elements :: "string \<Rightarrow> 'a \<Rightarrow> xml \<Rightarrow> (string,'a)emonad"
where "xml0elements reqName comb (XML name atts cs text) = 
  (if name = reqName \<and> atts = [] \<and> cs = [] \<and> text = None then return comb 
     else error (sConcat [''could not transform the following xml element (expected '',reqName,'')'',newline,sshow (XML name atts cs text)]))
    " 

fun xml1element :: "string \<Rightarrow> (xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> xml \<Rightarrow> (string,'b)emonad"
where "xml1element reqName p1 comb (XML name atts cs text) = 
  (if name = reqName \<and> atts = [] \<and> length cs = 1 \<and> text = None then
    (p1 (cs ! 0) \<guillemotright>= (return \<circ> comb))
  else error (sConcat [''could not transform the following xml element (expected '',reqName,'')'',newline,sshow (XML name atts cs text)]))
    " 

fun xml2elements :: "string \<Rightarrow> (xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'b)emonad) \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c) \<Rightarrow> xml \<Rightarrow> (string,'c)emonad"
where "xml2elements reqName p1 p2 comb (XML name atts cs text) = 
  (if name = reqName \<and> atts = [] \<and> length cs = 2 \<and> text = None then
    (p1 (cs ! 0) \<guillemotright>= (\<lambda> a. p2 (cs ! 1) \<guillemotright>= (return \<circ> comb a)))
  else error (sConcat [''could not transform the following xml element (expected '',reqName,'')'',newline,sshow (XML name atts cs text)]))
    "

fun xml3elements :: "string \<Rightarrow> (xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'b)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'c)emonad) \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'd) \<Rightarrow> xml \<Rightarrow> (string,'d)emonad"
where "xml3elements reqName p1 p2 p3 comb (XML name atts cs text) = 
  (if name = reqName \<and> atts = [] \<and> length cs = 3 \<and> text = None then
    (p1 (cs ! 0) \<guillemotright>= 
    (\<lambda> a. p2 (cs ! 1) \<guillemotright>= 
    (\<lambda> b. p3 (cs ! 2) 
    \<guillemotright>= (return \<circ> comb a b))))
  else error (sConcat [''could not transform the following xml element (expected '',reqName,'')'',newline,sshow (XML name atts cs text)]))
    "

fun xml4elements :: "string \<Rightarrow> (xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'b)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'c)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'d)emonad) \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'd \<Rightarrow> 'e) \<Rightarrow> xml \<Rightarrow> (string,'e)emonad"
where "xml4elements reqName p1 p2 p3 p4 comb (XML name atts cs text) = 
  (if name = reqName \<and> atts = [] \<and> length cs = 4 \<and> text = None then
    (p1 (cs ! 0) \<guillemotright>= 
    (\<lambda> a. p2 (cs ! 1) \<guillemotright>= 
    (\<lambda> b. p3 (cs ! 2) \<guillemotright>= 
    (\<lambda> c. p4 (cs ! 3) 
    \<guillemotright>= (return \<circ> comb a b c)))))
  else error (sConcat [''could not transform the following xml element (expected '',reqName,'')'',newline,sshow (XML name atts cs text)]))
    "
 
fun xml1to2elements :: "string \<Rightarrow> (xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'b)emonad) \<Rightarrow> ('a \<Rightarrow> 'b option \<Rightarrow> 'c) \<Rightarrow> xml \<Rightarrow> (string,'c)emonad"
where "xml1to2elements reqName p1 p2 comb (XML name atts cs text) = 
  (if name = reqName \<and> atts = [] \<and> length cs \<ge> 1 \<and> length cs \<le> 2 \<and> text = None then
    (p1 (cs ! 0) \<guillemotright>= (\<lambda> a. if (length cs = 2) then (p2 (cs ! 1) \<guillemotright>= (\<lambda> b. return (comb a (Some b)))) else return (comb a None)))
  else error (sConcat [''could not transform the following xml element (expected '',reqName,'')'',newline,sshow (XML name atts cs text)]))
    " 

fun xml2to3elements :: "string \<Rightarrow> (xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'b)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'c)emonad) \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c option \<Rightarrow> 'd) \<Rightarrow> xml \<Rightarrow> (string,'d)emonad"
where "xml2to3elements reqName p1 p2 p3 comb (XML name atts cs text) = 
  (if name = reqName \<and> atts = [] \<and> length cs \<ge> 2 \<and> length cs \<le> 3 \<and> text = None then
    (p1 (cs ! 0) \<guillemotright>= (\<lambda> a. p2 (cs ! 1) \<guillemotright>= (\<lambda> b. if (length cs = 3) then (p3 (cs ! 2) \<guillemotright>= (\<lambda> c. return (comb a b (Some c)))) else return (comb a b None))))
  else error (sConcat [''could not transform the following xml element (expected '',reqName,'')'',newline,sshow (XML name atts cs text)]))
    " 

fun xmlManyElements :: "string \<Rightarrow> (xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> ('a list \<Rightarrow> 'b) \<Rightarrow> xml \<Rightarrow> (string,'b)emonad"
where "xmlManyElements reqName p comb (XML name atts cs text) = 
  (if name = reqName \<and> atts = [] \<and> text = None then
     (mapMonad p cs \<guillemotright>= (return \<circ> comb))
   else error (sConcat [''could not transform the following xml element (expected '',reqName,'')'',newline,sshow (XML name atts cs text)]))"

fun xml1manyElements :: "string \<Rightarrow> (xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'b)emonad) \<Rightarrow> ('a \<Rightarrow> 'b list \<Rightarrow> 'c) \<Rightarrow> xml \<Rightarrow> (string,'c)emonad"
where "xml1manyElements reqName p1 p2 comb (XML name atts cs text) = 
  (if name = reqName \<and> atts = [] \<and> text = None \<and> cs \<noteq> [] then
     (p1 (hd cs) \<guillemotright>= (\<lambda> a. mapMonad p2 (tl cs) \<guillemotright>= (return \<circ> comb a)))
   else error (sConcat [''could not transform the following xml element (expected '',reqName,'')'',newline,sshow (XML name atts cs text)]))"

fun xml2manyElements :: "string \<Rightarrow> (xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'b)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'c)emonad) \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c list \<Rightarrow> 'd) \<Rightarrow> xml \<Rightarrow> (string,'d)emonad"
where "xml2manyElements reqName p1 p2 p3 comb (XML name atts cs text) = 
  (if name = reqName \<and> atts = [] \<and> text = None \<and> length cs \<ge> 2 then
     (p1 (cs ! 0) \<guillemotright>= (\<lambda> a. p2 (cs ! 1) \<guillemotright>= (\<lambda> b. mapMonad p3 (tl (tl cs)) \<guillemotright>= (return \<circ> comb a b))))
   else error (sConcat [''could not transform the following xml element (expected '',reqName,'')'',newline,sshow (XML name atts cs text)]))"

fun xmlOptionsHelp :: "(xml \<Rightarrow> (string,'a)emonad) list \<Rightarrow> string list \<Rightarrow> xml \<Rightarrow> (string,'a)emonad"
where "xmlOptionsHelp [] errs x = error (sConcat[''no option found for '',newline,sshow x] @ sConcat (map (\<lambda> err. sConcat [newline, ''-------'', newline, err])  errs ))"
   |  "xmlOptionsHelp (p # ps) errs x = (p x \<guillemotleft>= (\<lambda> err. xmlOptionsHelp ps (errs @ [err]) x))" 

fun xmlOptions :: "(xml \<Rightarrow> (string,'a)emonad) list \<Rightarrow> xml \<Rightarrow> (string,'a)emonad"
where "xmlOptions ps x = xmlOptionsHelp ps [] x"


fun xmlChange :: "(xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> (xml \<Rightarrow> (string,'b)emonad)"
where "xmlChange p f x = (p x \<guillemotright>= (\<lambda> a. return (f a)))"

fun xmlGetName :: "xml \<Rightarrow> string" 
where "xmlGetName (XML name _ _ _) = name"

fun xml2text :: "string \<Rightarrow> xml \<Rightarrow> (string,string)emonad"
where "xml2text reqName (XML name atts cs text) = 
  (if name = reqName \<and> atts = [] \<and> cs = [] \<and> text \<noteq> None \<and> the text \<noteq> [] then return (the text) else 
  error (sConcat [''could not extract text for '',reqName,'' from  '',newline,sshow (XML name atts cs text)]))"

fun digit2nat :: "char \<Rightarrow> (string,nat)emonad"
where "digit2nat x = (if x = CHR ''0'' then return 0 else
                     (if x = CHR ''1'' then return 1 else
                     (if x = CHR ''2'' then return 2 else
                     (if x = CHR ''3'' then return 3 else
                     (if x = CHR ''4'' then return 4 else
                     (if x = CHR ''5'' then return 5 else
                     (if x = CHR ''6'' then return 6 else
                     (if x = CHR ''7'' then return 7 else
                     (if x = CHR ''8'' then return 8 else
                     (if x = CHR ''9'' then return 9 else 
                       error ([x] @ '' is not a digit'')))))))))))"

fun string2natHelp :: "nat \<Rightarrow> nat endParser" 
where "string2natHelp n [] = return n"
    | "string2natHelp n (d # s) = (
      digit2nat d \<guillemotright>= (\<lambda> m. string2natHelp (10 * n + m) s))"

definition string2nat :: "nat endParser"
where "string2nat s = (if s = [] then error ''cannot convert empty string into number'' else string2natHelp 0 s)"

fun string2cint :: "cint endParser"
where "string2cint s = (case s of [] \<Rightarrow> error ''cannot convert empty string into number''
                            | h # t \<Rightarrow> let (pos,ss) = (if h = CHR ''-'' then (False,t) else (True,s))
                               in string2nat ss \<guillemotright>= (\<lambda> n. if pos \<or> n = 0 then return (Pos n) else return (Neg n)))"


fun xml2cint :: "string \<Rightarrow> xml \<Rightarrow> (string,cint)emonad"
where "xml2cint reqName x = (xml2text reqName x \<guillemotright>= string2cint)"

fun xml2nat :: "string \<Rightarrow> xml \<Rightarrow> (string,nat)emonad"
where "xml2nat reqName x = (xml2text reqName x \<guillemotright>= string2nat)"

fun xml2var :: "xml \<Rightarrow> (string,('a,string)term)emonad"
where "xml2var x = (xml2text ''var'' x \<guillemotright>= (return \<circ> Var))"

fun xml2name :: "xml \<Rightarrow> (string,string)emonad"
where "xml2name (XML name atts cs text) = 
     (if name = ''name'' \<and> cs = [] \<and> text \<noteq> None \<and> the text \<noteq> [] then
        (if atts = [] \<or> atts = [(''sharp'',''false'')] then return (the text)
           else error (''invalid attributes in '' +++ newline +++ sshow (XML name atts cs text)))
     else error (''could not parse name from '' +++ newline +++ sshow (XML name atts cs text)))" 

fun xml2sharpName :: "xml \<Rightarrow> (string,string shp)emonad"
where "xml2sharpName (XML name atts cs text) = 
     (if name = ''name'' \<and> cs = [] \<and> text \<noteq> None \<and> the text \<noteq> [] then
        (if atts = [] \<or> atts = [(''sharp'',''false'')] then return (Plain (the text))
           else (if atts = [(''sharp'',''true'')] then return (Sharp (the text))
             else error (''invalid attributes in '' +++ newline +++ sshow (XML name atts cs text))))
     else error (''could not parse sharp name from '' +++ newline +++ sshow (XML name atts cs text)))" 



function xml2term :: "xml \<Rightarrow> (string,(string,string)term)emonad"
where "xml2term x = 
  xmlOptions [xml2var, xml1manyElements ''funapp'' xml2name (xml1element ''arg'' xml2term id) Fun] x"  
by (pat_completeness, auto)

function xml2sharpTerm :: "xml \<Rightarrow> (string,(string shp,string)term)emonad"
where "xml2sharpTerm x = 
  xmlOptions [xml2var, xml1manyElements ''funapp'' xml2sharpName (xml1element ''arg'' xml2sharpTerm id) Fun] x"  
by (pat_completeness, auto)

function xml2context :: "xml \<Rightarrow> (string,(string,string)ctxt)emonad"
where "xml2context x = 
  xmlOptions [
       xml0elements ''box'' Hole,  
       xml4elements ''funContext'' 
          xml2name 
          (xmlManyElements ''before'' xml2term id)
          xml2context
          (xmlManyElements ''after'' xml2term id)
          More
     ] x"  
by (pat_completeness, auto)

definition xml2substitution :: "xml \<Rightarrow> (string,(string,string)subL)emonad"
where "xml2substitution = 
   xmlManyElements ''substitution''
      (xml2elements ''substEntry'' 
         xml2var
         xml2term
         Pair)
      id"

definition xml2rule :: "xml \<Rightarrow> (string,(string,string)rule)emonad"
where "xml2rule = xml2elements ''rule'' (xml1element ''lhs'' xml2term id) (xml1element ''rhs'' xml2term id) Pair" 

definition xml2rules :: "xml \<Rightarrow> (string,(string,string)trsL)emonad"
where "xml2rules = xmlManyElements ''rules'' xml2rule id"

definition xml2sharpRule :: "xml \<Rightarrow> (string,(string shp,string)rule)emonad"
where "xml2sharpRule = xml2elements ''rule'' (xml1element ''lhs'' xml2sharpTerm id) (xml1element ''rhs'' xml2sharpTerm id) Pair" 

definition xml2sharpRules :: "xml \<Rightarrow> (string,(string shp,string)trsL)emonad"
where "xml2sharpRules = xmlManyElements ''rules'' xml2sharpRule id"

definition xml2funcsym :: "xml \<Rightarrow> (string,string \<times> nat)emonad"
where "xml2funcsym = xml2elements ''funcsym'' (xml2text ''name'') (xml2nat ''arity'') Pair" 

definition xml2signature :: "xml \<Rightarrow> (string,(string \<times> nat)list)emonad"
where "xml2signature = xmlManyElements ''signature'' xml2funcsym id"

(* rules, signature, optional comment *)
datatype xtcTrs = XTCTRS "(string,string)trsL" "(string \<times> nat)list" "xml option" 

definition xml2xtcTrs :: "xml \<Rightarrow> (string,xtcTrs)emonad"
where "xml2xtcTrs = xml2to3elements ''trs'' xml2rules xml2signature 
   (\<lambda> x. if xmlGetName x = ''comment'' then return x else error 
     (sConcat [''unknown/unsupported third element in trs element'', newline, sshow x])) 
  XTCTRS" 



(* xtcTrs, startterm is given, elements from status metainformation ... which are ignored, attributes from problem before type, ... after type, header *)
datatype xtcProblem = XTCProblem xtcTrs "bool \<times> xml list \<times> (string \<times> string) list \<times> (string \<times> string) list \<times> string"

fun extractElement :: "'a \<Rightarrow> 'a list \<Rightarrow> (string,'a list \<times> 'a list)emonad"
where "extractElement _ [] = error ''not found''"
    | "extractElement a (b # bs) = (if a = b then return ([],bs)
                                     else extractElement a bs \<guillemotright>= (\<lambda> (bef,aft). return (b # bef, aft)))"

definition xmlStartTerm :: xml where
  "xmlStartTerm = XML ''startterm'' [] [XML ''full'' [] [] None] None"

(* TODO:
  to be on the safe side (for future extensions of xtc) 
  one may check that the rest elements are only
  status or metainformation
*)
fun xmldoc2xtcProblem :: "xmldoc \<Rightarrow> (string,xtcProblem)emonad"
where "xmldoc2xtcProblem (XMLDOC header xml) = (case xml of XML name atts cs text \<Rightarrow> 
  (if name = ''problem'' \<and> length cs \<ge> 2 \<and> length atts \<ge> 1 \<and> text = None then 
    (if cs ! 1 \<noteq> XML ''strategy'' [] [] (Some ''FULL'') 
      then error ''unsupported or unknown strategy'' 
         else extractElement (''type'',''termination'') atts \<leftarrow> (\<lambda> e. ''type termination attribute not found'')
            \<guillemotright>= (\<lambda> (bef,aft). xml2xtcTrs (cs ! 0) \<guillemotright>= (\<lambda> xtcTrs. 
              let (start,rest) = (if length cs > 2 \<and> cs ! 2 = xmlStartTerm 
                       then (True, List.drop 3 cs) else (False,List.drop 2 cs)) 
              in if ''startterm'' mem (map xmlGetName rest)
                     then error ''startterms unsupported'' else 
                     return (XTCProblem xtcTrs (start, rest, bef, aft, header)
          )))
    )
  else 
    error ''root node must be problem and have at least the two initial children trs and strategy''
  ))" 
  

fun parseXTCProblem :: "xtcProblem endParser"
where "parseXTCProblem s = (parseXMLDoc s \<guillemotright>= xmldoc2xtcProblem)"

fun fitLength :: "'a \<Rightarrow> nat \<Rightarrow> 'a list \<Rightarrow> 'a list"
where "fitLength _ 0 _ = []"
    | "fitLength a n [] = replicate n a"
    | "fitLength a (Suc n) (b # as) = b # (fitLength a n as)"

fun xml2redpair :: "(xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> xml \<Rightarrow> (string, 'a RedPair)emonad"
where "xml2redpair xml2sname = xmlOptions [
  xml1element ''redPair'' 
    (xml2manyElements ''interpretation'' 
       (xml1element ''type'' (xml0elements ''linearPolynomial'' ()) id)
       (xml1element ''domain'' (xml0elements ''naturals'' ()) id)
       (xml3elements ''interpret''
          xml2sname
          (xml2nat ''arity'')
          (xml1element ''function'' 
             (xml1manyElements ''linearPolynomial''
                (xml1element ''constant'' (xml2cint ''number'') id)
                (xml1element ''coefficient'' (xml2nat ''natural'') id)
                Pair)
             id)
          (\<lambda> f a poly. (f, (fst poly, fitLength 0 a (snd poly)))))
      (\<lambda> _ _ pi. pi))
    NegPolo
   ]"


function xml2dpProof :: "xml \<Rightarrow> (string,(string shp,string)DPProof)emonad"
where "xml2dpProof x = xmlOptions [ 
  xml0elements ''pIsEmpty'' 
     PisEmpty,
  xml3elements ''redPairProc'' 
     (xml2redpair xml2sharpName) 
     (xml1element ''dps'' xml2sharpRules id) 
     xml2dpProof 
     RedPairProc,
  xml4elements ''monoRedPairProc'' 
     (xml2redpair xml2sharpName) 
     (xml1element ''dps'' xml2sharpRules id) 
     (xml1element ''trs'' xml2sharpRules id) 
     xml2dpProof
     MonoRedPairProc,
  xmlManyElements ''depGraphProc''
     (xml1to2elements ''component'' 
         (xml1element ''dps'' xml2sharpRules id)
         xml2dpProof
         (\<lambda> dps prfOpt. (prfOpt, dps)))
     DepGraphProc    
  ] x"
by (pat_completeness, auto)

function xml2trsProof :: "xml \<Rightarrow> (string,(string,string)TRSProof)emonad"
where "xml2trsProof x = xmlOptions [
   xml0elements ''rIsEmpty''
      RisEmpty,
   xml2elements ''dpTrans'' 
      (xml1element ''dps'' xml2sharpRules id)
      xml2dpProof
      DPTrans,
   xml3elements ''ruleRemoval''
      (xml2redpair xml2name) 
      (xml1element ''trs'' xml2rules id)
      xml2trsProof
      RuleRemoval
  ] x" 
by (pat_completeness, auto)

definition xml2trsDisproof :: "xml \<Rightarrow> (string,(string,string)TRSDisproof)emonad"
where "xml2trsDisproof = 
   xml2manyElements ''loop'' 
      xml2substitution
      xml2context
      xml2term 
      (\<lambda> \<sigma> C ts. Loop (ts, C, \<sigma>))" 


fun xmldoc2proof :: "xmldoc \<Rightarrow> (string,((string,string)TRSProof, (string,string)TRSDisproof)either)emonad"
where "xmldoc2proof (XMLDOC header (XML name atts cs text)) = 
  (if name = ''proof'' \<and> length cs = 1 \<and> text = None then 
    xmlOptions [xmlChange xml2trsProof Left, xmlChange xml2trsDisproof Right] (cs ! 0)
    else error ''invalid root element, expecting proof with exactly one child and no text'')"

fun parseProof :: "((string,string)TRSProof, (string,string)TRSDisproof) either endParser"
where "parseProof s = (parseXMLDoc s \<guillemotright>= xmldoc2proof)"

declare parseProof.simps[simp del]
declare parseXTCProblem.simps[simp del]

(* ... to xml *)

class xml = type + 
  fixes xml :: "'a \<Rightarrow> xml"

instantiation "term" :: (sshow,sshow)xml
begin
fun xml_term 
  where "xml_term (Var x) = XML ''var'' [] [] (Some (sshow x))" 
      | "xml_term (Fun f ts) = XML ''funapp'' [] 
              (XML ''name'' [] [] (Some (sshow f)) # 
              (map (\<lambda> t. XML ''arg'' [] [xml_term t] None) ts)) 
           None"
instance ..
end


instantiation xtcTrs :: xml
begin
fun xml_sig :: "(string \<times> nat)list \<Rightarrow> xml"
where "xml_sig fas = XML ''signature'' [] (map (\<lambda> (f,a). XML ''funcsym'' [] [
         XML ''name'' [] [] (Some f),
         XML ''arity'' [] [] (Some (sshow a)) 
         ] None) fas) None"
fun xml_rules :: "(string,string)trsL \<Rightarrow> xml"
where "xml_rules lrs = XML ''rules'' [] (map (\<lambda> (l,r). XML ''rule'' [] [
         XML ''lhs'' [] [xml l] None,         
         XML ''rhs'' [] [xml r] None
         ] None) lrs) None"
fun xml_xtcTrs :: "xtcTrs \<Rightarrow> xml"
where "xml_xtcTrs (XTCTRS rls sig comment) = 
     XML ''trs'' [] ([xml_rules rls,xml_sig sig] @ (case comment of None \<Rightarrow> [] | Some c \<Rightarrow> [c])) None" 
instance ..
end

instantiation xtcProblem :: xml
begin
fun xml_xtcProblem :: "xtcProblem \<Rightarrow> xml"
where "xml_xtcProblem (XTCProblem xtcTrs (start, ssm, bef, aft, header)) = 
   XML ''problem'' (bef @ [(''type'',''termination'')] @ aft) ([
     xml xtcTrs,
     XML ''strategy'' [] [] (Some ''FULL'')
   ] @ (if start then [xmlStartTerm] else []) @ ssm) None"
instance ..
end

fun xtc2xmldoc :: "xtcProblem \<Rightarrow> xmldoc"
where "xtc2xmldoc (XTCProblem xtcTrs (start,ssm,bef, aft, header)) = 
          XMLDOC header (xml (XTCProblem xtcTrs (start,ssm, bef, aft, header)))"

instantiation xtcProblem :: sshow
begin
fun sshow_xtcProblem 
where "sshow_xtcProblem p = sshow (xtc2xmldoc p)"
instance ..
end


end